# ais/core.py
import numpy as np


def ais_predict(data, labels, detector_count=10, split_ratio=0.8):
    split = int(split_ratio * len(data))
    X_train, X_test = data[:split], data[split:]
    y_train, y_test = labels[:split], labels[split:]

    idx = np.random.choice(len(X_train), detector_count, replace=False)
    detectors = X_train[idx]
    detector_labels = y_train[idx]

    predictions = []
    for x in X_test:
        i = np.argmin(np.linalg.norm(detectors - x, axis=1))
        predictions.append(detector_labels[i])

    acc = np.mean(predictions == y_test)
    return acc



import random
from deap import base, creator, tools, algorithms


def run_ga():
    creator.create("FitnessMax", base.Fitness, weights=(1.0,))
    creator.create("Individual", list, fitness=creator.FitnessMax)

    toolbox = base.Toolbox()
    toolbox.register("bit", random.randint, 0, 1)
    toolbox.register(
        "individual", tools.initRepeat, creator.Individual, toolbox.bit, 10
    )
    toolbox.register("population", tools.initRepeat, list, toolbox.individual)
    toolbox.register("evaluate", lambda ind: (sum(ind),))
    toolbox.register("mate", tools.cxTwoPoint)
    toolbox.register("mutate", tools.mutFlipBit, indpb=0.1)
    toolbox.register("select", tools.selTournament, tournsize=3)

    pop = toolbox.population(n=20)
    algorithms.eaSimple(pop, toolbox, cxpb=0.5, mutpb=0.2, ngen=10, verbose=False)

    best = tools.selBest(pop, 1)[0]
    return best, sum(best)



from setuptools import setup, find_packages

setup(
    name="ais-toolkit-rohan",
    version="0.1.0",
    description="AIS, Genetic Algorithm, and Ant Colony Optimization toolkit",
    author="Rohan Kaitake",
    author_email="your_email@example.com",
    packages=find_packages(),
    install_requires=[
        "numpy",
        "deap",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



# ANT COLONY OPTIMIZATION
import numpy as np
import random


class AntColony:
    def __init__(
        self, dist_matrix, n_ants, n_best, n_iterations, decay, alpha=1, beta=2
    ):
        self.dist_matrix = dist_matrix
        self.pheromone = np.ones(dist_matrix.shape) / len(dist_matrix)
        self.all_inds = range(len(dist_matrix))
        self.n_ants = n_ants
        self.n_best = n_best
        self.n_iterations = n_iterations
        self.decay = decay
        self.alpha = alpha
        self.beta = beta

    def run(self):
        shortest_path = None
        all_time_shortest_path = ("placeholder", np.inf)
        for _ in range(self.n_iterations):
            all_paths = self.gen_all_paths()
            self.spread_pheronome(all_paths, self.n_best)
            shortest_path = min(all_paths, key=lambda x: x[1])
            if shortest_path[1] < all_time_shortest_path[1]:
                all_time_shortest_path = shortest_path
            self.pheromone *= 1 - self.decay
        return all_time_shortest_path

    def spread_pheronome(self, all_paths, n_best):
        sorted_paths = sorted(all_paths, key=lambda x: x[1])
        for path, dist in sorted_paths[:n_best]:
            for move in path:
                self.pheromone[move] += 1.0 / self.dist_matrix[move]

    def gen_path_dist(self, path):
        total_dist = 0
        for ele in path:
            total_dist += self.dist_matrix[ele]
        return total_dist

    def gen_all_paths(self):
        all_paths = []
        for _ in range(self.n_ants):
            path = self.gen_path(0)
            all_paths.append((path, self.path_distance(path)))
        return all_paths

    def gen_path(self, start):
        path = []
        visited = set()
        visited.add(start)
        prev = start
        for _ in range(len(self.dist_matrix) - 1):
            move = self.pick_move(self.pheromone[prev], self.dist_matrix[prev], visited)
            path.append((prev, move))
            prev = move
            visited.add(move)
        path.append((prev, start))  # return to start
        return path

    def pick_move(self, pheromone, dist, visited):
        pheromone = np.copy(pheromone)
        pheromone[list(visited)] = 0

        row = pheromone**self.alpha * ((1.0 / dist) ** self.beta)
        norm_row = row / row.sum()
        move = np.random.choice(self.all_inds, 1, p=norm_row)[0]
        return move

    def path_distance(self, path):
        total = 0
        for ele in path:
            total += self.dist_matrix[ele]
        return total


def run_aco():
    # Sample distance matrix (you can modify this)
    cities = 5
    np.random.seed(42)
    dist_matrix = np.random.randint(1, 100, size=(cities, cities))
    np.fill_diagonal(dist_matrix, 999999)  # large number to prevent loops

    aco = AntColony(dist_matrix, n_ants=10, n_best=5, n_iterations=50, decay=0.1)
    path, distance = aco.run()
    return path, distance
